(* \comment{ *)
module P = Parser
module B = Bunch

type tag = Term.tag = Add | Mul

type term = Term.term  =
    Bvar of string
  | Evar of tag*string*int*string
  | Const of string
  | Appl of string * (term Bunch.bunch)
type varbind = Term.varbind = Varbind of string
type goal = P.goal
type clause = P.clause

type tflag = int
type eflag = On | Off

type listused = 
    Locked 
  | Open of clause B.bunch * eflag
  | Full of clause B.bunch * clause B.bunch
type context = Epsilon | Leavesome

exception Finished
exception End
exception Failed of clause B.bunch

open P

let mtrue, atrue = Const "I", Const "T"

(* --------------------------------------------------------------------- *)
(* The only unit actually used is gmunit = multiplicative unit as a goal *)
let cmunit = Catom mtrue
let caunit = Catom atrue
let gmunit = Gatom mtrue
let gaunit = Gatom atrue
let emptyc = Catom (Const "")
let i = gmunit

(* end comment } *)
(* ----------------------------------------------------------------- *)
(* {\bf Define some operators to improve readability} *)
let (--) a b = B.subtract b a
let (&) = B.ajoin 
let ( * ) = B.mjoin 

(* \comment{ begins a section commented out in the thesis  *)
(* ----------------------------------------------------------------- *)
(* Printing and debugging functions   *)

let debug = ref false

let setdebug t = debug := t

let print p g ctxt = if not !debug then () else
  let s = match g with
      Gatom a -> Term.printTerm ctxt a
    | _ -> printGoal ctxt g
  and pc = printClause ctxt
  in
      B.print_bunch pc p;
      print_string (" |- "^s^"\n")

let print_context ctxt = 
  let f (x,y) = "("^(Term.printTerm1 ctxt x)
    ^ ","
    ^(Term.printTerm1 ctxt y)
    ^ ")"
  in B.print_bunch f ctxt; print_string "\n"

let print1 a b ctxt = ()
(*  print_context ctxt;
  print_string ("Unified: "^Term.printTerm ctxt a^" with "
                   ^Term.printTerm ctxt b^"\n")
*)

exception Found

let backtrack s = () (*; print_string (s^"\n")*)
(* proof := List.tl !proof; print_string ("bt in "^s^"\n")*)

(* ----------------------------------------------------------------- *)
(* * Left and & Left rules, to be made at the beginning    *)

let rec explode bunch =
  let f = function
      Cmand (c1,c2) -> 
          (explode (B.L c1)) * (explode (B.L c2))
    | Caand (c1,c2) -> 
          (explode (B.L c1)) & (explode (B.L c2))
    | c -> B.L c
  in 
      B.proc f bunch

(* ----------------------------------------------------------------- *)
(* Create logic variables with a unique name    *)
let tag = ref ""

let next_tag () = 
   let rec f n = 
      let a = Char.code 'a' in
      let range = Char.code 'z' - a + 1
      in
      !tag.[n] <- Char.chr(((Char.code (!tag.[n])-a + 1) mod range) + a);
      if !tag.[n] = 'a' then 
      if n = 0 then (tag := !tag ^ "a")
      else f (n-1)
      else ()
   in let n = String.length !tag in
       if n = 0 then tag := "a" else f(n-1);
       String.copy !tag

let next_evar = function
    Evar(t,s,n,b) -> Evar(t,s,n+1,b)
  | _ -> raise (Failure "in next_evar()")


let newAddvar(s) = Evar(Add,next_tag(),1,s)
let newMulvar(s) = Evar(Mul,next_tag(),1,s)


(*------------------------------------------------------------------*)
(* Substitutions and the like  *)

let shadow (Varbind v1) (Varbind v2) = (v1 = v2)

let formsubst t x = 
  let rec gsb = function
      Gatom a -> Gatom (Term.subst t x a)
    | Gcommand a -> Gcommand (Term.subst t x a)
    | Gaand(g1,g2) -> Gaand(gsb g1, gsb g2)
    | Gmand(g1,g2) -> Gmand(gsb g1, gsb g2)
    | Gaor(g1,g2) -> Gaor(gsb g1, gsb g2)
    | Gmimp(c,g,f) -> Gmimp(csb c, gsb g,f)
    | Gaimp(c,g) -> Gaimp(csb c, gsb g)
    | Gex(v,g) -> Gex(v,if shadow x v then g else gsb g)
    | Gexnew(v,g) -> Gexnew(v,if shadow x v then g else gsb g)
    | Gall(v,g) -> Gall(v,if shadow x v then g else gsb g)
    | Gallnew(v,g) -> Gallnew(v,if shadow x v then g else gsb g)
  and csb = function
      Catom a -> Catom (Term.subst t x a)
    | Caand(c1,c2) -> Caand(csb c1, csb c2)
    | Cmand(c1,c2) -> Cmand(csb c1, csb c2)
    | Cmimp(g,a,n) -> Cmimp(gsb g, Term.subst t x a,n)
    | Caimp(g,a) -> Caimp(gsb g, Term.subst t x a)
    | Call(v,c) -> Call(v,if shadow x v then c else csb c)
    | Callnew(v,c) -> Callnew(v,if shadow x v then c else csb c)
  in
      (gsb, csb)

let gsubst t x = (fst (formsubst t x))

let csubst t x = (snd (formsubst t x))

(* ----------------------------------------------------------------- *)
(* Implementation of the CutAxiom rule   *)

let rec axiom a = function
    B.Empty -> B.Empty
  | (B.L c) as bunch -> if c == a then B.Empty else bunch
  | B.MB (h::[]) ->
        axiom a h
  | B.MB (h::tl) ->
        let ntl = 
          match tl with 
             [x] -> x
            | x1 -> B.MB x1
        in
        if B.mem a h then 
            (axiom a h) * ntl
        else h * (axiom a ntl)
  | B.AB (h::[]) ->
        axiom a h
  | (B.AB (h::tl)) as bunch ->
        if not (B.mem a bunch) then bunch else
        if B.mem a h then axiom a h
        else axiom a (B.AB tl)
  | _ -> raise B.BunchErrorException


(* ----------------------------------------------------------------- *)

let printclause = P.printClause B.Empty

let readgoal = P.read_goal

let isModule = 
  let test s = match s.[0] with
      'A'..'Z' -> true
    | _ -> false
  in function
      Bvar s -> false
    | Evar _ -> false
    | Const s -> test s
    | Appl(s,b) -> test s

let rec map f c = function
    B.Empty -> raise (Failure "empty in map (interpreter)")
  | B.L a -> f a c
  | B.MB (m::[]) -> map f c m 
  | B.MB (m::tl) -> map f (map f c (B.MB tl)) m
  | B.AB (m::[]) -> map f c m 
  | B.AB (m::tl) -> map f (map f c (B.AB tl)) m
  | _ -> raise (Failure "something strange in map (interpreter)")

let filesubst pars vars clauses= 
  let b = B.bzip pars vars in
  let f (p,v) c = csubst p v c in
      B.bmap (fun c -> map (fun z -> f z) c b) clauses

let rec mklist_of_vars = function
    B.Empty -> []
  | B.L (Varbind a) -> [a]
  | B.MB list -> List.concat (List.map mklist_of_vars list)
  | B.AB list -> List.concat (List.map mklist_of_vars list)

let readfile s = 
        let file = P.openFile s in
        let vars = P.readVariables file in
            P.readClauses (mklist_of_vars vars) file 

let loadModule = function
    Bvar s -> raise (Failure "bound variable in load module")
  | Evar _ -> raise (Failure "logic variable in load module")
  | Const s -> readfile (s^".bi")
  | Appl(s,b) -> 
        let file = P.openFile (s^".bi") in
        let vars = P.readVariables file in
        let clauses = P.readClauses (mklist_of_vars vars) file in
            filesubst b vars clauses

let rec write ctxt = function
    [] -> ()
  | (x::tl) -> let t = Term.findbyname x ctxt in
        print_string (x^":"^(Term.printTerm ctxt t))


let executeCommand ctxt = function
    Const "exit" -> raise End
  | Const "nl" ->  print_string "\n"
  | Const "sp" ->  print_string " "
  | Appl("write", b) -> 
        let f t = print_string ((Term.printTerm ctxt t)^" ")
        in B.biter f b
  | s -> raise (Failure ("Unknown keyword "^(Term.printTerm ctxt s)))

let rec load = function
    Catom a as phi -> 
        if isModule a then loadModule a
        else explode (B.L phi)
  | Caand(phi,psi) -> load phi & load psi
  | Cmand(phi,psi) -> load phi * load psi
  | s -> explode (B.L s)

(* ----------------------------------------------------------------- *)
(* This is what used to be called solve   *)

(* end of code comented out in the thesis }  *)
let rec (|-) ((gamma,s)as program) goal epsilon n sigma sc = 
  print gamma goal sigma;
  match goal with
(* \end{code} \hrule \vspace{5pt}
   When an atom is found we can call a left rule, which is done in the
   function \ttt{|--} below. We need to pass on the flags
   (\eepsilon\ and $n$), the substitution (\ssigma) and the 
   success continuation (\ttt{sc}).
\vspace{-8pt} \begin{code} *)
    | Gatom alpha -> (program |-- alpha) epsilon n sigma sc
(* \end{code} \hrule \vspace{5pt}
   If the goal is a command we execute it. The current substitution
   may be needed for side effects, so we pass it on.
\vspace{-8pt} \begin{code} *)
    | Gcommand s -> executeCommand sigma s; 
          sc program epsilon n sigma 
(* \end{code} \hrule \vspace{5pt}
   For $\n R$ there are five subcases. The first three arise when
   \eepsilon\ $\neq \emptybunch$ and the last two when \eepsilon\ $= 
   \emptybunch$. However, notice that in all cases
   $\pair{\Gamma|\stack}$ is used in the left subproof. This means
   that it is possible to take common factor with respect to the left
   branch, and analyse cases after this has been done.
\vspace{-3pt} \begin{code} *)
    | Gaand (phi, psi) -> 
          (program |- phi) epsilon n sigma 
(* \end{code}
   Note that it is all right to ignore the \eepsilon\ flag, since it is 
   not going to be changed. 
\vspace{-3pt} \begin{code} *)
              (fun (delta, s') _ n' sigma' ->
(* \end{code} 
   First, if \eepsilon\ $\neq \emptyflag$ then we are in one of the
   first three cases:
\vspace{-3pt} \begin{code} *)
                  if epsilon <> Epsilon then (* \n R, \n Ri, \n Rii *)
                      if n' = n then (* \n R *)
(* \end{code} 
   The rule for $\n R$ is
   \[
   \ptb \mng nn{\Gamma|\stack}{\Delta|\stack}\phi \quad
        \mng 0m{\Gamma - \Delta|[]}{\emptyflag|[]}\psi 
   \using \n R
   \ptj \mng nn{\Gamma|\stack}{\Delta|\stack}{\phi\n\psi}
   \pte
   \]

   In the continuation we can ignore the program (we know that it has to
   be $\pair{\emptybunch,[]}$) and the flags \eepsilon\ and $m$ from the 
   right branch.
\vspace{-3pt} \begin{code} *)
                          ((gamma -- delta , []) |- psi) Epsilon 0 sigma'
                              (fun _ _ _ -> sc (delta,s') epsilon n)
(* \end{code}
   At this point we know that $n'\neq n$. It has to be greater, then,
   since the only way of reducing $n$ is through the $\Unit^i$ rule,
   which is not applicable since \eepsilon\ $\neq \emptyflag$. Now the
   distinction between $\n R^i$ and $\n R^{ii}$ is based on whether
   $T$ appears in a goal position in $\phi$ or not. So we
   unconditionally do the left branch and decide which rule to use
   after testing the $T$ counter.
\vspace{-3pt} \begin{code} *)
                      else if n' > n then(* \n Ri, \n Rii *)
                          (program |- psi) epsilon n sigma'
(* \end{code}
   In the following continuation we deliberately ignore the stack
   returned, as well as the \eepsilon\ flag.
\vspace{-3pt} \begin{code} *)
                              (fun (delta',_) _ n'' -> 
(* \end{code}
   The rule for $\n R^i$ is
   \[
   \ptb \manager{\Gamma|\stack}{\Delta|\stack}\phi \quad
        \mng nn{\Gamma|\stack}{\Delta'|\stack}\psi 
   \using \n R^i(n'>n {\rm\ and\ } \Delta'\subb\Delta)
   \ptj \mng nn{\Gamma|\stack}{\Delta'|\stack}{\phi\n\psi}
   \pte
   \]
   For it to succeed, we need to test whether the condition
   $\Delta'\subb\Delta$ holds. 
\vspace{-3pt} \begin{code} *)
                                  if n'' = n && B.subbunch delta' delta (* \n Ri *)
                                  then  sc (delta',s') epsilon n
(* \end{code}
   The rule for $\n R^{ii}$ is
   \[
   \ptb \manager{\Gamma|\stack}{\Delta|\stack} \phi \quad
        \mng n{n''}{\Gamma|\stack}{\Delta'|\stack} \psi 
   \using \n R^{ii}(n',n''>n)
   \ptj \manager{\Gamma|\stack}{\Delta\cap\Delta'|\stack}{\phi\n\psi}
   \pte
   \]
   In this case we succeed, but taking the intersection of both
   remainders. This produces backtracking on the right subproof,
   increasing the non-determinism of the program. For a more detailed
   discussion of this issue see \S\ref{non-determinism} in 
   \chapterref{opsemantics} 
\vspace{-3pt} \begin{code} *)
                                  else if n'' > n (* \n Rii *)
                                  then sc (B.intersection delta delta',s') 
                                          epsilon n'
                                  else raise (Failure "strange n in \n Rii"))
                      else raise (Failure "strange n in \n R ")
(* \end{code}
   The other two cases, $\n R^{iii}$ and $\n R^{iv}$, arise when
   \eepsilon\ $= \emptyflag$. We distinguish the cases by looking in
   the stack returned by the proof of $\phi$. It is convenient to look 
   first at the case $\n R^{iv}$.
\vspace{-3pt} \begin{code} *)
                  else match s' with (* \eepsilon\ \ensuremath{= \emptyflag} *)
(* \end{code}
   The rule for $\n R^{iv}$ is 
   \[ 
   \ptb \manager{\xi;\Gamma|\opene{\xi}}
                {\emptyflag|\full{\Delta,\xi}}\phi \quad
        \mng 0{n''}{\xi;(\Gamma - \Delta)|[]}
                   {\emptyflag|[]}\psi
   \using \n R^{iv}
   \ptj \mng n{n'''}{\xi;\Gamma|\opene{\xi}}{\emptyflag|\full{\Delta,\xi}}
                         \phi\n\psi
   \pte
   \]

\vspace{-3pt} \begin{code} *)
                      (Full(delta,xsi))::l -> (* \n  Riv *)
                          ((xsi&(gamma--delta),[]) |- psi) Epsilon 0
                              sigma' (fun (empty, _ ) _ n''->
                                  sc (empty,s') Epsilon 
                                      (if n' > n && n'' > n then n' else n))
(* \end{code}
   Finally, the rule for $\n R^{iii}$ is
   \[
   \ptb \manager{\Gamma|\stack}{\emptyflag|\stack}\phi \quad
        \mng n{n''}{\Gamma|\stack}{\emptyflag|\stack'}\psi 
   \using \n R^{iii}
   \ptj \mng n{n'''}{\Gamma|\stack}{\emptyflag|\stack'}{\phi\n\psi}
   \pte
   \]
   and we use it if the stack is of any other shape. 
\vspace{-3pt} \begin{code} *)
                    | _ ->   (* \n Riii *)
                          (program |- psi) Epsilon n sigma'
                              (fun np _ n'' ->
                                  sc np Epsilon 
                                      (if n' > n && n'' > n then (n+1) else n)
                                      ))
(* \end{code} \hrule \vspace{5pt}
   The rule for $* R$ is
   \[
   \ptb \manager{\Gamma|\push{\locked}}{\Gamma'|\push{\locked}}
                            \phi  \quad
   \managerppp{\Gamma'|\stack}{\Delta|\stack'}\psi
   \using * R
   \ptj \managerpp{\Gamma|\stack}{\Delta|\stack'} {\phi*\psi}
   \pte
   \]
 \begin{code}*)
    | Gmand(phi,psi) -> 
          ((gamma, Locked::s) |- phi) Leavesome n sigma
              (fun (gamma',Locked::s) _ n' sigma' ->
                  ((gamma',s) |- psi) epsilon n' sigma' sc) 
(* \end{code} \hrule \vspace{5pt} The rules for $\/ R$ are
   \[
   \ptb \manager{\Gamma|\stack}{\Delta|\stack'} \phi 
   \ptj \manager{\Gamma|\stack}{\Delta|\stack'} {\phi \/ \psi}
   \using \/ R
   \pte
   \qquad
   \ptb \manager{\Gamma|\stack}{\Delta|\stack'} \psi
   \ptj \manager{\Gamma|\stack}{\Delta|\stack'} {\phi \/ \psi}
   \using \/ R
   \pte
   \]
   \noindent so we first try $\phi$ and in case of failure (return) we
   try $\psi$ \begin{code}*)
    | Gaor(phi,psi) ->
          (program |- phi) epsilon n sigma sc;
          (program |- psi) epsilon n sigma sc
(* \end{code} \hrule \vspace{5pt} 
   The rule for $-* R$ is
   \[
   \ptb \manager{\phi,\Gamma|\stack}{\Delta|\stack'}\psi
   \using -* R 
   \ptj \manager{\Gamma|\stack}{\Delta'|\stack'}{\phi-*\psi}
   \pte
   \]
   \begin{code}*)
    | Gmimp(phi,psi,f) -> 
          let exphi = explode (load phi) in
	  let newprog = if f then (gamma * exphi,s) else (exphi * gamma,s) in
              (newprog |- psi) epsilon n sigma
                  (fun (delta,s') epsilon' n' ns -> 
                      if n' > n 
                      then sc (delta--exphi,s') epsilon' n' ns
                      else if (B.check exphi delta)
                      then sc (delta,s') epsilon' n' ns)

(* \end{code} \hrule \vspace{5pt}
   There are various rules for $-> R$. 
   \vspace{-8pt} \begin{code} *)
    | Gaimp(phi, psi) ->
          let exphi = explode (load phi)
          in
          let lst = match s with
(* \end{code} \hrule \vspace{5pt}
   \[
   \ptb \manager{\phi;\Gamma|[]}{\emptyflag|[]}\psi
   \using -> R
   \ptj \manager{\Gamma|[]}{\emptyflag|[]}{\phi->\psi}
   \pte
   \]
\vspace{-8pt} \begin{code} *)
            | [] -> []
(* \end{code}
\vspace{-8pt} \begin{code} *)

            | (Open (a,tick))::l -> Open (exphi & a,tick)::l
(* \end{code}
   \[
   \ptb \manager{\phi;\Gamma|\opene{\phi}}
                {\emptyflag|\full{\Delta,\phi}}\psi
   \using -> R^i
   \ptj \manager{\Gamma|\push{\locked}}
                {\Delta|\push{\locked}}{\phi->\psi}
   \pte
   \]
\vspace{-8pt} \begin{code} *)
            | Locked ::l -> (Open (exphi,On))::l
            | _ ::l -> raise (Failure "Weird s in Gaimp")
          in
              ((exphi & gamma,lst) |- psi) Epsilon n sigma
                  (fun (delta,l) na tf -> 
                      let cont, nt = match l with
                          [] -> delta -- exphi, tf
                        | (Full (gamma',_ ))::l -> gamma' -- exphi, tf
                              | (Open (a,On)):: l -> gamma, tf + 1
                        | _ -> delta -- exphi, tf
                      in
                          sc (cont,s) epsilon nt)
(* \end{code} \hrule \vspace{5pt}
\vspace{-8pt} \begin{code} *)
    | Gex ((Varbind v as vr), g) ->
          let var = newAddvar(v) 
          in let nc = (B.L (Const v,var)) & sigma in
              (program |- gsubst var vr g) epsilon n nc sc
(* \end{code} \hrule \vspace{5pt}
\vspace{-8pt} \begin{code} *)
    | Gexnew ((Varbind v as vr), g) ->
          let var = newMulvar(v)
          in let nc = (B.L (Const v,var)) * sigma in
              (program |- gsubst var vr g) epsilon n nc sc
    | Gall (v, g) -> raise (Failure "Forall in goals not implemented yet")
    | Gallnew (v, g) -> raise (Failure "Allnew in goals not implemented yet")

(* and what used to be called match_atom    *)

and (|--) (gamma,s) alpha epsilon n sigma sc = 
  let rec test c = 
    let r t = 
      match (t,Term.dereference alpha sigma) with 
          (Const a, Const b) -> if a = b then c else emptyc
        | (Appl(a,_), Appl(b,_)) -> if a = b then c else emptyc
        | _ -> emptyc
    in
            match c with
            Catom t1  -> r t1
          | Cmimp(g,t1,n) -> r t1
          | Caimp(g,t1) -> r t1
          | Call (v,cl) -> 
                if test cl = emptyc then emptyc else c
          | Callnew (v,cl) ->
                if test cl = emptyc then emptyc else c
          | Cmand(c1,c2) ->
                if test c1 = emptyc && test c2 = emptyc then emptyc else c
          | Caand(c1,c2) ->
                if test c1 = emptyc && test c2 = emptyc then emptyc else c
  and resolve = function
      B.Empty -> emptyc
    | B.L c -> test c
    | B.MB (h::[]) -> 
          resolve h
    | B.MB (h::tl) ->
          let g = resolve h in
              if g = emptyc then resolve (B.MB tl)
              else g
    | B.AB (h::[]) -> 
          resolve h
    | B.AB (h::tl) ->
          let g = resolve h in
              if g = emptyc then resolve (B.AB tl)
              else g
    | _ -> raise B.BunchErrorException
  in let c = ref (resolve gamma)
  and b = ref gamma in
      while !c <> emptyc do
          let gamma' = axiom !c gamma in
          let tfound a = 
            if a = atrue then 1 else 0
          in
          let clause = 
            match !c with 
                      Call (_,_) | Callnew (_,_) -> strip !c
              | _ -> !c
          in
              begin match clause with
(* \end{code} \hrule \vspace{5pt}
   There are two rules for $\CutAxiom$. One of them applies just when the 
   top of the stack has got an open box with the theorem flag and the atom 
   considered is not a member of the content of the box ($\xi$)
   \[
   \ptb \manager{\Gamma'|\stack}{\Delta|\stack'}I
   \using \CutAxiom^\dagger
   \ptj \manager{\Gamma(\alpha)|\stack}{\Delta|\stack'}\alpha
   \pte
   \qquad
   \ptb \manager{\Gamma'|\open{\xi}}{\Delta|\stack'}I
   \using \CutAxiom^{i\dagger}
   \ptj \manager{\xi;\Gamma(\alpha)|\opene{\xi}}{\Delta|\stack'}\alpha
   \pte
   \]

\vspace{-8pt} \begin{code} *)
                | Catom alpha' -> 
                      Term.unify alpha' alpha (fun sigma' ->
                          print1 alpha alpha' sigma';
                          let s' = 
                            match s with
                                Open (xi,On)::l ->
                                    if not (B.mem clause xi) then 
                                        Open (xi,Off)::l
                                    else s
                              | _ -> s
                          in
                          ((gamma',s') |- i) epsilon 
                              (n + (tfound alpha'))
                              sigma' sc) sigma
(* \end{code} \hrule \vspace{5pt}
   The rule for $-* L$ is
   \[
   \ptb \manager{\Gamma'|\stack}{\Delta|\stack'} \phi 
   \using -* L ^\dagger
   \ptj \manager{\Gamma(\phi -* \alpha)|\stack}{\Delta|\stack'}\alpha 
   \pte 
   \]
\vspace{-8pt} \begin{code} *)
                | Cmimp(phi, alpha', k) ->
                      let gamma' = 
                        if k = 0 then B.mnw !c gamma
                        else if k = 1 then gamma'
                        else B.replace (B.L(reduce !c))
                                !c (B.mnw !c gamma)
                      in
                      Term.unify alpha' alpha (fun sigma' ->
                          print1 alpha alpha' sigma';
                          ((gamma',s) |- phi) epsilon 
                              (n + (tfound alpha'))
                              sigma' sc) sigma 
(* \end{code} \hrule \vspace{5pt}
   \[
   \ptb \mng 0m{\Theta;\phi->\alpha|[]}{\emptyflag|[]}\phi 
   \qquad 
   \manager{\Gamma'|\stack}{\Delta|\stack'}I 
   \using -> L^\dagger
   \ptj \manager{\Gamma(\Theta;\phi -> \alpha)|\stack}{\Delta|\stack'}
                          \alpha 
   \pte
   \]

\vspace{-8pt} \begin{code} *)
                | Caimp(phi, alpha') ->
                      let theta = B.getabunch !c gamma in
                          Term.unify alpha' alpha
                              (fun sigma' -> 
                                  print1 alpha alpha' sigma';
                                  ((theta,[]) |- phi) Epsilon 0 sigma'
                                      (fun _ _ _ sigma'' -> 
                                          ((gamma',s) |- i) epsilon n 
                                              sigma'' sc))
                              sigma 
(* \end{code} \hrule \vspace{5pt}
   We can still get a multiplicative or additive conjunction if it was
   wrapped in a quantifier

\vspace{-8pt} \begin{code} *)
                | Cmand(_,_) | Caand(_,_) ->
                      let program = B.replace (explode (B.L clause)) !c gamma
                      in
                          ((program,s) |-- alpha) epsilon n sigma sc
                | _ -> backtrack ""
              end;
              b := B.remfirst !c !b;
              c := resolve !b
      done;
      match alpha with
(* \end{code} \hrule \vspace{5pt}
   The rule for $\TUnit$ is
   \[
   \ptb
   \using \TUnit
   \ptj \mng n{n+1}{\Gamma|\stack}{\Gamma|\stack}T
   \pte
   \]
\vspace{-8pt} \begin{code} *)
        | Const "T" -> sc (gamma,s) epsilon (n + 1) sigma
(* \end{code} \hrule \vspace{5pt}
   

\vspace{-8pt} \begin{code} *)
        | Const "I" -> (* gmunit *)
              if epsilon = Epsilon && (gamma <> B.Empty)                  
(* \end{code}
   If $n = 0$ we use the rule for $\Unit^{ii}$ 
   \[
   \ptb
   \using \Unit^{ii}
   \ptj \managere{\phi;\Gamma|\opene{\phi}}
                 {\emptyflag|\full{\Gamma,\phi}}I
   \pte
   \]
\vspace{-8pt} \begin{code} *)
              then if n = 0 then match s with
                | (Open (phi,tick))::l ->
                      sc (B.Empty,(Full (gamma,phi)::l)) Epsilon 0 sigma
                | _ -> ()
(* \end{code}
   Else, we use the rule $\Unit^i$
   \[
   \ptb
   \using \Unit^i (n> 0)
   \ptj \mng n0{\Gamma|\stack}{\emptyflag|\stack}I
   \pte
   \]
\vspace{-8pt} \begin{code} *)
              else (* n > 0 *)
                  sc (B.Empty,s) Epsilon 0 sigma
(* \end{code} \hrule \vspace{5pt}
   Finally, if $\emptyflag$ is not present or the bunch is empty, 
   there is no reason not to succeed, and we use the rule $\Unit$
   \[
   \ptb
   \using \Unit
   \ptj \mng 00{\Gamma|\stack}{\Gamma|\stack}I
   \pte
   \]
\vspace{-8pt} \begin{code} *)
              else sc (gamma,s) epsilon n sigma
(* \end{code} \hrule \vspace{5pt}
   If none of these cases applies we backtrack, but we should never reach 
   this case
\vspace{-8pt} \begin{code} *)
        | _ -> backtrack ""
and strip = function
    Call ((Varbind s) as v,cl) -> csubst (newAddvar(s)) v (strip cl)
  | Callnew ((Varbind s) as v,cl) -> csubst (newMulvar(s)) v (strip cl)
  | cl -> cl
and reduce = function
    Call (v,cl) -> Call (v,reduce cl)
  | Callnew (v,cl) -> Callnew (v,reduce cl)
  | Cmimp (phi, alpha, n)-> Cmimp(phi, alpha, n-1)
  | cl -> cl

(* \comment{ *)

(* ----------------------------------------------------------------- *)
(* The external interface   *)

let rec getvars = function
    Gex (Varbind v, g) -> B.L (Term.Const v) & getvars g
  | Gexnew (Varbind v, g) ->  B.L (Term.Const v) * getvars g
  | _ -> B.Empty

let prove bc g = 
  tag := "";
  ((explode bc,[])|- g) Epsilon 0 B.Empty 
      (fun (leftover,s) epsilon n subst -> 
          Term.printSubst (getvars g) subst;
          if epsilon = Epsilon && (leftover <> B.Empty) && n = 0 then
              begin print_string "failed with left over: "; 
              B.print_bunch (P.printClause subst) leftover;
              print_string "\n**********\n";
              raise Finished
              end
          else 
              begin
              print_string "yes.\n**********\n";
              let s = read_line() in 
                  if s = ";" then () 
                  else raise Finished
   end);
  print_string "no.\n**********\n";
  raise Finished

(*finalcomment*)
(* end of skipped code } *)

