type symbol = 
    EOF 
  | Bang
  | Comma | Semicolon (* comma and semicolon *)
  | Dot | Equals
  | Minus | Plus
  | Munit | Aunit
  | Bottom
  | Mimp | Aimp (* implications *)
  | RevMimp | RevAimp (*reverse implications *)
  | Mand | Aand (* multiplicative and additive and *)
  | Aor (* additive or *)
  | Mnot | Anot
  | OpenRound | CloseRound (* nesting of bunches 
			      or forcing associativity  *)
  | OpenSquare | CloseSquare  (* forall *)
  | OpenDoubleSquare | CloseDoubleSquare  (* forall new *)
  | OpenAngle | CloseAngle  (* exists *)
  | OpenDoubleAngle | CloseDoubleAngle  (* exists new *)
  | OpenCurly | CloseCurly   (* nothing yet  *)
  | Int of int | Id of string (* numbers, identifiers *)
  | Keyword of string

let keywords = ["exit"; "write"; "nl"; "sp"; "wsubst"; "import" ]

let line_number = ref 1

let error s = 
  print_string("("^(string_of_int !line_number)^")Error: "^s)

let rec spaces s = 
  match s with parser
      [< '' '|'\t' >] -> spaces s
    | [< ''\n'>] -> line_number := !line_number+1; spaces s
    | [< >] -> s

let rec eat_rest_line = parser
    | [< ''\n'; s>] -> line_number := !line_number+1; s
    | [< 'c; s >] -> eat_rest_line s

let int_of_digit = function
    '0'..'9' as c -> (int_of_string (Char.escaped c)) - 
	    (int_of_string "0")
  | _ -> raise (Failure "Not a digit")

let rec integer n = parser
      [< ''0'..'9' as c ; s>] -> integer (10*n + int_of_digit c) s
    | [< >] -> n

let string_of_char = function
    'a'..'z'|'A'..'Z'|'_'|'0'..'9' as c -> Char.escaped c
  | _ -> raise (Failure "Not a letter")

let rec var_name n = parser
      [< ''a'..'z'|'A'..'Z'|'_'|'0'..'9'  as c ; s >] -> 
	  var_name (n^string_of_char c) s
    | [< >] -> n

let rec quoted str = parser
    | [< ''\\' ; s>] -> 
	  begin match s with parser
	      [< ''n'>] -> quoted (str^"\n") s
	  end 
    | [<''"' >] -> str  (*    *)
    | [< 'c ; s>] -> quoted (str^Char.escaped c) s
    | [< >] -> raise (Failure "Oh dear, what have we done?")

let rec lexer s = match (spaces s) with parser
    [< '';' >] -> [< 'Semicolon ; lexer s>]
  | [< ''#' >] -> [< lexer (eat_rest_line s)>] (* comments *)
  | [< ''!' >] -> [< 'Bang ; lexer s>]
  | [< '',' >] -> [< 'Comma ; lexer s>]
  | [< ''.' >] -> [< 'Dot ; lexer s>]
  | [< ''=' >] -> [< 'Equals ; lexer s>]
  | [< ''+' >] -> [< 'Plus ; lexer s>]
  | [< ''-' >] -> 
	begin match s with parser
	    [< ''*'>] -> [< 'Mimp ; lexer s>]
	  | [< ''>'>] -> [< 'Aimp ; lexer s>]
	  | [< >] -> [< 'Minus ; lexer s>]
	end
  | [< ''*'>] -> 
	begin match s with parser
	    [< ''-' >] -> [< 'RevMimp ; lexer s>]
	  | [< >] -> [< 'Mand ; lexer s>]
	end
  | [< ''&' >] -> [< 'Aand ; lexer s>]
  | [< ''|' >] -> [< 'Aor ; lexer s>]
  | [< ''_' >] -> [< 'Bottom ; lexer s >]
  | [< ''~' >] -> [< 'Mnot ; lexer s >]
  | [< ''(' >] -> [< 'OpenRound ; lexer s>]
  | [< '')' >] -> [< 'CloseRound ; lexer s>]
  | [< ''[' >] -> 
	begin match s with parser
	    [< ''[' >] -> [< 'OpenDoubleSquare ; lexer s>]
	  | [< >] -> [< 'OpenSquare ; lexer s>]
	end
  | [< '']' >] -> 
	begin match s with parser
	    [< '']'>] -> [< 'CloseDoubleSquare ; lexer s>]
          | [< >] -> [< 'CloseSquare ; lexer s>]
	end
  | [< ''<' >] -> 
	begin match s with parser
	    [< ''<'>] -> [< 'OpenDoubleAngle ; lexer s>]
	  | [< ''-'>] -> [< 'RevAimp ; lexer s>]
	  | [< >] -> [< 'OpenAngle ; lexer s>]
	end
  | [< ''>' >] -> 
	begin match s with parser
	    [< ''>'>] -> [< 'CloseDoubleAngle ; lexer s>]
	  | [< >] -> [< 'CloseAngle ; lexer s>]
	end
  | [< ''{'>] -> [< 'OpenCurly ; lexer s>]
  | [< ''}'>] -> [< 'CloseCurly ; lexer s>]
  | [< ''"'>] -> [< 'Id (quoted "" s) ; lexer s>]
(*  | [< ''"'>] -> [< 'Id (quoted (string_of_char(Stream.next s)) s) ; lexer s>] " *)
  | [< ''0'..'9' as c >] -> 
    [< 'Int (integer (int_of_digit c) s); lexer s>]
  | [< ''a'..'z' | 'A'..'Z' as c >] ->
    let variable = var_name (string_of_char c) s in
    let sym = 
    if List.mem variable keywords then Keyword variable
    else 
    match variable with
      "I" -> Munit
    | "T" -> Aunit
    | "not" -> Anot
    | str -> Id str 
    in
       [< 'sym ; lexer s>]
  | [< >] -> [< 'EOF >]

let sym_to_string = function
    EOF -> "EOF "
  | Bang -> "!"
  | Comma -> ","
  | Semicolon -> ";"
  | Dot -> "."
  | Equals -> "="
  | Minus -> "-"
  | Plus -> "+"
  | Munit -> "I"
  | Aunit -> "T"
  | Bottom -> "_"
  | Mnot -> "~"
  | Anot -> " not "
  | Mimp -> "-*"
  | Aimp -> "->"
  | RevMimp -> "*-"
  | RevAimp -> "<-"
  | Mand -> "*"
  | Aand -> "&"
  | Aor -> "|"
  | OpenRound -> "("
  | CloseRound -> ")"
  | OpenSquare -> "["
  | CloseSquare -> "]"
  | OpenDoubleSquare -> "[["
  | CloseDoubleSquare -> "]]"
  | OpenAngle -> "<"
  | CloseAngle -> ">"
  | OpenDoubleAngle -> "<<"
  | CloseDoubleAngle -> ">>"
  | OpenCurly -> "{"
  | CloseCurly -> "}"
  | Int n -> "N "^(string_of_int n)^" "
  | Id s -> "Id: "^s
  | Keyword s -> "Kw: "^s
