
module L = Lexer
module T = Term

type symbol = L.symbol

type tag = Add | Mul

type term = T.term
type varbind = T.varbind 

type clause = 
    Catom of term
  | Caand of clause * clause
  | Cmand of clause * clause
  | Cmimp of goal * term * int
  | Caimp of goal * term
  | Call of varbind * clause
  | Callnew of varbind * clause
and goal = 
    Gatom of term 
  | Gaand of goal * goal
  | Gmand of goal * goal
  | Gaor of goal * goal
  | Gmimp of clause * goal * bool
  | Gaimp of clause * goal
  | Gex of varbind * goal
  | Gexnew of varbind * goal
  | Gall of varbind * goal
  | Gallnew of varbind * goal
  | Gcommand of term

type cbunch = clause Bunch.bunch
type substitution = (term*term)Bunch.bunch

type temp = 
    Atom of term 
  | Aand of temp * temp
  | Mand of temp * temp
  | Aor of temp * temp
  | Mimp of temp * temp * int 
  | Aimp of temp * temp
  | Ex of varbind * temp
  | Exnew of varbind * temp
  | All of varbind * temp
  | Allnew of varbind * temp
  | Command of term

let bottom = T.Const "_"

let strip = function
    Catom a -> a
  | _ -> raise (Failure "not atomic")

let rec  tempToClause = function
    Atom a -> Catom a
  | Aand(t1,t2) -> Caand (tempToClause t1, tempToClause t2)
  | Mand(t1,t2) -> Cmand (tempToClause t1, tempToClause t2)
  | Aimp(t1,Atom a) -> Caimp(tempToGoal t1, a)
  | Mimp(t1,Atom a, n) -> Cmimp(tempToGoal t1, a, n)
  | All (v, t) -> Call(v,tempToClause t)
  | Allnew (v, t) -> Callnew(v,tempToClause t)
  | _  -> raise (Failure ("cannot convert temp into clause"))
and tempToGoal = function
    Atom a -> Gatom a
  | Command a -> Gcommand a
  | Aand(t1,t2) -> Gaand (tempToGoal t1, tempToGoal t2)
  | Mand(t1,t2) -> Gmand (tempToGoal t1, tempToGoal t2)
  | Aimp(t1,t2) -> Gaimp(tempToClause t1, tempToGoal t2)
  | Mimp(t1,t2,n) -> Gmimp(tempToClause t1, tempToGoal t2,if n = 0 then true else false)
  | Aor (t1,t2) -> Gaor(tempToGoal t1, tempToGoal t2)
  | All (v, t) -> Gall(v,tempToGoal t)
  | Allnew (v, t) -> Gallnew(v,tempToGoal t)
  | Ex (v, t) -> Gex (v, tempToGoal t)
  | Exnew (v, t) -> Gexnew (v, tempToGoal t)


let rec readTemp bv = parser
    [< 'L.OpenRound; e = readTemp bv ;
	'L.CloseRound;
	e1 = readRestTemp bv e >] -> e1
  | [< 'L.OpenSquare; 
	e = readA bv L.CloseSquare (fun x y -> All(x,y)) >] -> e
  | [< 'L.OpenDoubleSquare; 
	e = readA bv L.CloseDoubleSquare (fun x y -> Allnew(x,y))>] -> e
  | [< 'L.OpenAngle; 
	e = readE bv L.CloseAngle (fun x y -> Ex(x,y)) >] -> e
  | [< 'L.OpenDoubleAngle; 
	e = readE bv L.CloseDoubleAngle (fun x y -> Exnew(x,y))>] -> e
  | [< 'L.Keyword s; t = T.readTerm bv s;
	e = readRestTemp bv (Command t) >] -> e
  | [< 'L.Id s; t = T.readTerm bv s ; e = readRestTemp bv (Atom t) >] -> e
  | [< 'L.Munit; e = readRestTemp bv (Atom (T.Const "I")) >] -> e
  | [< 'L.Aunit; e = readRestTemp bv (Atom (T.Const "T")) >] -> e
  | [< 'L.Bottom; e = readRestTemp bv (Atom bottom) >] -> e
  | [< 'L.Mnot; e = readTemp bv ; 
	e1 = readRestTemp bv (Mimp(e,Atom bottom, 1)) >] -> e1
  | [< 'L.Anot; e = readTemp bv ; 
	e1 = readRestTemp bv (Aimp(e,Atom bottom)) >] -> e1
and readRestTemp bv e1 =  parser
    [< 'L.Mand;  e = readTemp bv ; 
	e2 = readRestTemp bv (Mand(e1, e)) >] -> e2
  | [< 'L.RevMimp ; s >] ->
	begin match s with parser
	    [< 'L.Int n; e = readTemp bv ; 
		e2 = readRestTemp bv (Mimp(e, e1, n)) >] -> e2
	  | [< e = readTemp bv ;
		e2 = readRestTemp bv (Mimp(e, e1, 1)) >] -> e2
	end
  | [< 'L.RevAimp; e = readTemp bv ; 
	e2 = readRestTemp bv (Aimp(e, e1)) >] -> e2
  | [< 'L.Aand;  e = readTemp bv ; 
	e2 = readRestTemp bv (Aand(e1, e)) >] -> e2
  | [< 'L.Mimp ; s>] -> 
	begin match s with parser
	    [< 'L.Int n; e = readTemp bv ;
		e2 = readRestTemp bv (Mimp(e1, e, n)) >] -> e2
	  | [< e = readTemp bv ;
		e2 = readRestTemp bv (Mimp(e1, e, 1)) >] -> e2
	end
  | [< 'L.Aimp; e = readTemp bv ;
	e2 = readRestTemp bv (Aimp(e1, e)) >] -> e2
  | [< 'L.Aor;  e = readTemp bv ; 
	e2 = readRestTemp bv (Aor(e1, e)) >] -> e2
  | [< >] -> e1
and readA bv close constr = parser
    [< 'L.Id v; e = carryonA v bv close constr>] -> e
  | [< 'close; e = readTemp bv >] -> e
  | [< >] -> raise (Failure "reading variables")
and carryonA v bv close constr = parser
    [< 'L.Comma; e = readA (v::bv) close constr >] -> constr (T.Varbind v) e
  | [< 'close; e = readTemp (v::bv) >] -> constr (T.Varbind v) e
and readE bv close constr = parser
    [< 'L.Id v; e = carryonE v bv close constr>] -> e
  | [< 'close; e = readTemp bv >] -> e
  | [< >] -> raise (Failure "reading variables")
and carryonE v bv close constr = parser
    [< 'L.Comma; e = readE (v::bv) close constr >] -> constr (T.Varbind v) e
  | [< 'close; e = readTemp (v::bv) >] -> constr (T.Varbind v) e

let readGoal bv = parser
    [< e = readTemp bv >] -> tempToGoal e

let readClause bv = parser
    [< e = readTemp bv >] -> tempToClause e

let readVars = parser
    [< 'L.Id v >] -> T.Varbind v

let rec mkbunchparser f = parser
    [< 'L.OpenRound ; s >] -> [< 'Bunch.OBracket ; mkbunchparser f s >]
  | [< 'L.CloseRound ; s >] -> [< 'Bunch.CBracket ; mkbunchparser f s >]
  | [< 'L.Comma ; s >] -> [< 'Bunch.Mul ; mkbunchparser f s >]
  | [< 'L.Semicolon ; s >] -> [< 'Bunch.Add ; mkbunchparser f s >]
  | [< 'L.Dot >] -> [< 'Bunch.End >]
  | [< e = f ; s>] -> [< 'Bunch.Leaf e ; mkbunchparser f s >]
  | [<  >] -> [< 'Bunch.End  >]

let mkparser lst s = mkbunchparser (readClause lst) s

let readClauses lst s = Bunch.build_bunch (mkparser lst s)

let readVariables s = Bunch.build_bunch (mkbunchparser readVars s)

let openFile s = L.lexer(Stream.of_channel (open_in s))

let read_clauses s = readClauses [] (L.lexer(Stream.of_string s))

let read_goal s = readGoal [] (L.lexer(Stream.of_string s))

let rec printClause ctxt t = 
  let ps = print_string in
  match t with
      Catom a -> T.printTerm ctxt  a
    | Caand(m, n) ->"("^(printClause ctxt m)^ " & " ^(printClause ctxt n)^")"
    | Cmand(m, n) -> "("^(printClause ctxt m)^ " * " ^(printClause ctxt n)^")"
    | Cmimp(g, (T.Const "_"), _) ->"~"^(printGoal ctxt g)
    | Cmimp(g, a, n) ->
	  let s = if n = 1 then " " else 
	  if n = 0 then "00 " else (string_of_int n)^" " in
	      "("^(printGoal ctxt g)^" -*"^s^(T.printTerm ctxt a)^")"
    | Caimp(g, (T.Const "_")) -> "not "^(printGoal ctxt g)
    | Caimp(g, a) -> "("^(printGoal ctxt g)^" -> "^(T.printTerm ctxt a)^")"
    | Call ((T.Varbind s), c) -> "["^s^"]"^(printClause ctxt c)
    | Callnew ((T.Varbind s), c) -> "[["^s^"]]"^(printClause ctxt c)
and printGoal ctxt t =
  match t with
      Gatom a -> T.printTerm ctxt a
    | Gcommand t -> T.printTerm ctxt t
    | Gaand(m, n) -> "("^(printGoal ctxt m)^" & "^(printGoal ctxt n)^")"
    | Gmand(m, n) -> "("^(printGoal ctxt m)^" * "^(printGoal ctxt n)^")"
    | Gaor(m, n) -> "("^(printGoal ctxt m)^" | "^(printGoal ctxt n)^")"
    | Gmimp(c, Gatom (T.Const "_"),f) -> "~"^(printClause ctxt c)
    | Gaimp(c, Gatom (T.Const "_")) -> "not "^(printClause ctxt c)
    | Gmimp(c, g, f) -> 
	  let s = if f then "0 " else " " in
	  "("^(printClause ctxt c)^" -*"^s^(printGoal ctxt g)^")"
    | Gaimp(c, g) -> "("^(printClause ctxt c)^" -> "^(printGoal ctxt g)^")"
    | Gex ((T.Varbind s), g) -> "<"^s^">"^(printGoal ctxt g)
    | Gexnew ((T.Varbind s), g) -> "<<"^s^">>"^(printGoal ctxt g)
    | Gall ((T.Varbind s), g) -> "["^s^"]"^(printGoal ctxt g)
    | Gallnew ((T.Varbind s), g) -> "[["^s^"]]"^(printGoal ctxt g)

