module L = Lexer
module B = Bunch

type tag = Add | Mul

type name = string

type term = 
    Bvar of string
  | Evar of tag*string*int*name
  | Const of string
  | Appl of string * (term B.bunch)
and varbind = Varbind of string

type substitution = (term*term) B.bunch

let rec readTerm bv s = parser
    [< e = readBunch bv >] ->
	let flag = List.mem s bv in
	    if e = B.Empty then
		if flag then
		    Bvar s
		else Const s
	    else if flag then raise (Failure "Quantifying predicate")
	    else Appl (s,e)
and readBunch bv p = match p with parser
    [< 'L.OpenRound >] -> 
	B.build_bunch [< 'B.OBracket ; mktermparser 1 bv p >]
  | [< >] -> B.Empty
and mktermparser n bv p = match p with parser
    [< 'L.OpenRound >] -> [< 'B.OBracket ; mktermparser (n+1) bv p >]
  | [< 'L.CloseRound >] -> 
	if n = 1 then [< 'B.CBracket ; 'B.End >]
	else [< 'B.CBracket ; mktermparser (n-1) bv p >]
  | [< 'L.Comma >] -> [< 'B.Mul ; mktermparser n bv p >]
  | [< 'L.Semicolon >] -> [< 'B.Add ; mktermparser n bv p >]
  | [< 'L.Id s >] -> 
	[< 'B.Leaf (readTerm bv s p) ; mktermparser n bv p >]
  | [<  >] -> raise (Failure "wrong shape of bunch")

let rec subst s ((Varbind v) as vb) = function
    (Bvar bvname) as c -> if v = bvname then s else c
  | Appl(t, b) -> Appl(t, B.bmap (subst s vb) b)
  | c -> c

exception Found of term

let lookup e subst =
  match e with
      (Evar(t,a,n,_)) ->
	  let f = function
	      (Evar(t1,a1,n1,_),term) -> 
		  if t1=t & a1=a & n1=n then raise (Found term)
	    | _ -> ()
	  and s = ref None in 
	      begin try B.biter f subst 
	      with Found tr -> s := Some tr
	      end;
	      !s
    | _ -> raise (Failure "not an Evar in lookup")

let dereference s subst =
    let rec deref  = function
	((Evar _) as s) ->
	    begin match(lookup s subst) with
	  	None -> s
	      | Some s0 -> deref s0
	    end
      | (Appl(s0, b)) -> Appl(s0, B.bmap deref b)
      | s -> s
    in deref s

let project_substitution evars subst = 
  let rec prc acc_subst = function
      B.Empty -> acc_subst
    | B.L (bdevar,t) -> 
	  if B.bexists (fun evar-> evar = bdevar) evars then
	      (bdevar, dereference t subst)::acc_subst
	  else  acc_subst
    | B.AB (h::[]) -> prc acc_subst h
    | B.AB (h::tl) -> prc (prc acc_subst h) (B.AB tl)
    | B.MB (h::[]) -> prc acc_subst h
    | B.MB (h::tl) -> prc (prc acc_subst h) (B.MB tl)
    | _ -> raise (Failure "bad bunch structure")
  in prc [] subst 

let findbyname t ctxt = 
  let f (x,y) = 
    match x with
	Evar (_,_,_,n) | Const n  -> 
	    if n = t then raise (Found (dereference y ctxt))
      | _ -> ()
  in 
      let r = ref (Const t) in
	  begin try B.biter f ctxt
	  with Found term -> r := term
	  end;
	  !r
		  

let rec unify s t sc subst =
  let rec bunify a b sf sb= 
    match (a,b) with
	(B.Empty,B.Empty) -> sf sb
      | (B.L c1, B.L c2) -> 
	    unify c1 c2 sf sb
      | (B.AB (h1::[]),B.AB (h2::[])) ->
	    bunify h1 h2 sf sb
      | (B.AB (h1::tl1),B.AB (h2::tl2)) ->
	    bunify h1 h2 
		(fun ns -> bunify (B.AB tl1)(B.AB tl2) 
			sf ns) sb
      | (B.MB (h1::[]),B.MB (h2::[])) ->
	    bunify h1 h2 sf sb
      | (B.MB (h1::tl1),B.MB (h2::tl2)) ->
	    bunify h1 h2 
		(fun ns -> bunify (B.MB tl1)(B.MB tl2) 
			sf ns) sb
      | ( _ , _ ) -> ()
  in
      match (s,t) with
	  ((Evar _) as s,  t)  -> unify_evar s t sc subst
    	| (s, ((Evar _) as t)) -> unify_evar t s sc subst
    	| (Const c1, Const c2) -> if c1 = c2 then sc subst else ()
    	| (Appl(t1,b1),Appl(t2,b2)) -> 
	      if t1 = t2 then bunify b1 b2 sc subst
    	| _ -> ()

and unify_evar e t sc subst =
  match lookup e subst with
      None -> sc (B.ajoin (B.L (e,t)) subst)
    | Some s0 -> unify s0 t sc subst

let rec printTerm ctxt = function
    Bvar s -> s
  | Evar (t,s,n,a) as e -> 
	begin match lookup e ctxt with
	    Some lae -> 
	    	if lae = e then
	    	    s^ if t = Add then "!" else ""
	    	else printTerm ctxt lae
	  | _ -> s ^ if t = Add then "!" else ""
	end
  | Const s -> s
  | Appl (s, tb) ->
	begin
	match tb with
	    B.L _ -> s^"("^ B.bunch_to_string (printTerm ctxt) tb ^")"
	  | _ -> s ^ B.bunch_to_string (printTerm ctxt) tb
	end

let rec printTerm1 ctxt = function
    Bvar s -> s
  | Evar (t,s,n,s1) as e -> 
	s^ (if t = Add then "!:" else ":") ^ s1 ^ " " ^ string_of_int n
  | Const s -> s
  | Appl (s, tb) ->
	begin
	match tb with
	    B.L _ -> s^"("^ B.bunch_to_string (printTerm1 ctxt) tb ^")"
	  | _ -> s ^ B.bunch_to_string (printTerm1 ctxt) tb
	end

let printSubst evars subst =
  let f (x,y) = print_string ((printTerm subst x) ^ " <- " ^ 
				 printTerm subst y ^ "\n")
  in
      List.iter f (project_substitution evars subst)
